  export PYTHONPATH=`pwd`/src:/Users/kumykov/workspace/hub-rest-api-python 
  export BD_URL=https://ec2-44-202-54-168.compute-1.amazonaws.com
  export API_TOKEN=MTJlMmI5MGMtNmI0MS00ZGJiLThkMWEtMzNjNDNjMGU5ODAzOjc5Y2Q1NzdlLTNmYzEtNDMzNS1iZDYwLTY1NmQ3YTdkNGFlZA==
